<?php

namespace Drupal\payson_checkout\PluginForm\OffsiteRedirect;

use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\PluginForm\PaymentOffsiteForm as BasePaymentOffsiteForm;
use Drupal\commerce_payment\Entity\Payment;
use Drupal\payson_checkout\PaysonHandler;
use Drupal\Core\Form\FormStateInterface;
use GuzzleHttp\Client;
use Drupal\Core\Url;

use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsNotificationsInterface;

class PaysonCheckoutPaymentForm extends BasePaymentOffsiteForm {

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $this->form = parent::buildConfigurationForm($form, $form_state);

    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = $this->entity;
    /** @var \Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayInterface $this->payment_gateway_plugin */
    $this->payment_gateway_plugin = $this->entity->getPaymentGateway()->getPlugin();
    /** @var \Drupal\commerce_order\Entity\OrderInterface $order */
    $order = $payment->getOrder();
    $this->configuration = $this->payment_gateway_plugin->getConfiguration();

    $paysonHandler = new PaysonHandler($this->payment_gateway_plugin);

    if (!$order->getTotalPrice()->isPositive()) {
      throw new \Exception('Order ID "' . $payment->getOrderId() . '": Total price is not a positive amount.');
    }

    $storeUrl = \Drupal::request()->getSchemeAndHttpHost();

    // check if we already have a valid payment/checkout on the order and reuse it. Otherwise create new.
    $paysonCheckoutObj = $this->getPaymentFromOrder($order);

    // "Payson script url finder"
    // print_r(var_dump($paysonCheckoutObj->snippet));

    // Create div in form that will be replaced with payson checkout form via JS.
    $this->form['payson'] = [
      '#type' => ['markup'],
      '#prefix' => "<div class='payson-form'>",
      '#suffix' => '</div>',
      '#markup' => $paysonCheckoutObj->snippet,
    ];

    // and include the js.
    if($this->configuration['mode'] == 'live') {
      $this->form['#attached']['library'][] = 'payson_checkout/checkout_live';
    }else {
      $this->form['#attached']['library'][] = 'payson_checkout/checkout_dev';
    }

    // save the payment so we can verify it against the ipn from payson after the payment.
    $payment = \Drupal\commerce_payment\Entity\Payment::create([
      'type' => 'payment_default',
      'payment_gateway' => $order->get('payment_gateway')->entity->get('id'),
      'order_id' => $order->id(),
      'amount' => $order->getTotalPrice(),
      'remote_id' => $paysonCheckoutObj->id,
      'remote_state' => $paysonCheckoutObj->status,
      'expire' => $paysonCheckoutObj->expirationTime,
      'state' => 'new',
    ]);
    $payment->save();

    // no need to call buildRedirectForm as we are embeding an iframe.
    return $this->form;
  }

  public function checkoutPostData($order) {
    $storeName = \Drupal::entityTypeManager()
      ->getStorage('commerce_store')
      ->load($order->getStoreId())
      ->getName();
    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = $this->entity;

    $checkout = [];

    $checkout['merchant'] = [
      'checkoutUri' => $this->form['#cancel_url'],
      'confirmationUri' => $this->form['#return_url'],
      'notificationUri' => $this->payment_gateway_plugin->getNotifyUrl()->toString() . '?order_id=' . $payment->getOrderId(),
      'termsUri' => $this->configuration['terms'],
      'reference' => t('Order id ' . $payment->getOrderId() . ' at ' . $storeName)->render(),
    ];

    $checkout['order'] = [
      'currency' => $order->getTotalPrice()->getCurrencyCode(),
      'items' => $this->orderItems($order),
      'totalPriceIncludingTax' => $order->getTotalPrice()->getNumber(),
    ];

    $checkout['gui'] = [
      'colorScheme' => $this->configuration['theme'],
      'requestPhone' => $this->requestPhone()[0],
      'phoneOptional' => $this->requestPhone()[1],
      'verification' => $this->configuration['verification'],
    ];

    $checkout['customer'] = $this->getCustomer($order);

    return $checkout;
  }

  public function orderItems($order) {
    $order_items = $order->getItems();
    $checkoutItems = [];
    // add all orderitems to the checkout.
    foreach($order->getItems() AS $row => $order_item) {
      foreach($order_item->getAdjustments() AS $adjustment) {
        if ($adjustment->getType() == 'tax' && $adjustment->isIncluded()) {
          // tax and it is included in the price.
          $checkoutItems[$row]['unitPrice'] = $order_item->getUnitPrice()->getNumber();
          $checkoutItems[$row]['taxRate'] = $adjustment->getPercentage();
        } elseif($adjustment->getType() == 'tax' && !$adjustment->isIncluded()) {
          // tax that is not included in the price.
          $checkoutItems[$row]['unitPrice'] = $order_item->getUnitPrice()->getNumber() + $adjustment->getAmount()->getNumber();
          $checkoutItems[$row]['taxRate'] = $adjustment->getPercentage();
        }
      }
      $checkoutItems[$row]['name'] = $order_item->getTitle();
      $checkoutItems[$row]['quantity'] = $order_item->getQuantity();
      $checkoutItems[$row]['reference'] = $order_item->getPurchasedEntity()->getSku();
      $checkoutItems[$row]['type'] = 'physical';
    }
    $adjustmentRow = count($checkoutItems) -1; // array start with zero and count doesn't ...

    // now add all the adjustments on the order items. Discounts and fees and so on.
    foreach($order->getItems() AS $row => $order_item) {
      // fix all adjustments that isn't tax.
      foreach($order_item->getAdjustments() AS $adjustment) {
        if ($adjustment->getType() == 'tax') {
          // we took tax in the previous loop.
          continue;
        }
        if ( $adjustment->isIncluded()) {
          // we dont need to do anything if it is included in the price.
          continue;
        }
        $adjustmentRow++;
        switch ($adjustment->getType()) {
          case 'promotion':
            $checkoutItems[$adjustmentRow]['type'] = 'discount';
            break;
          case 'fee':
          case 'custom':
          default:
            if ($adjustment->getAmount()->getNumber() < 0 ) {
              // if negative set as discount otherwise set it as fee.
              $checkoutItems[$adjustmentRow]['type'] = 'discount';
            } else {
              $checkoutItems[$adjustmentRow]['type'] = 'fee';
            }
            break;
        }
        $checkoutItems[$adjustmentRow]['unitPrice'] = $adjustment->getAmount()->getNumber();
        $checkoutItems[$adjustmentRow]['name'] = $adjustment->getLabel();
        $checkoutItems[$adjustmentRow]['quantity'] = $order_item->getQuantity();
      }
    }

    $adjustmentRow = count($checkoutItems) -1; // array start with zero and count doesn't ...
    // order adjustments
    foreach($order->getAdjustments() AS $row => $orderAdjustment) {
      $adjustmentRow++;
      if (!$orderAdjustment->isIncluded()) {
        // included is already included in total so we dont need to care about them
        switch( $orderAdjustment->getType() ) {
          case 'fee':
          case 'custom':
          case 'promotion':
          case 'tax':
          default:
            if ($orderAdjustment->getAmount()->getNumber() < 0 ) {
              // if negative set as discount otherwise set it as fee.
              $checkoutItems[$adjustmentRow]['type'] = 'discount';
            } else {
              $checkoutItems[$adjustmentRow]['type'] = 'fee';
            }
          break;
        }
        $checkoutItems[$adjustmentRow]['unitPrice'] = $orderAdjustment->getAmount()->getNumber();
        $checkoutItems[$adjustmentRow]['name'] = $orderAdjustment->getLabel();
        $checkoutItems[$adjustmentRow]['quantity'] = 1;
     }
    }
    return $checkoutItems;
  }

  public function requestPhone() {
    // first is request phone, second is required.
    switch ($this->configuration['phone_required']) {
      case 0:
        // skip phone.
        return [0, 0];
      case 1:
        // phone is required.
        return [1, 1];
      case 2:
        // phone is optional.
        return [1, 0];
    }
    
  }
  
  public function getCustomer($order) {
    $billingAddress = $order->getBillingProfile()->get('address')->first();
    return [
      'city' => $billingAddress->getLocality(),
      'countryCode' => $billingAddress->getCountryCode(),
      'email' => $order->getEmail(),
      'firstName' => $billingAddress->getGivenName(),
      'lastName' => $billingAddress->getFamilyName(),
      'phone' => '',
      'postalCode' => $billingAddress->getPostalCode(),
      'street' => $billingAddress->getAddressLine1(),
    ];
  }

 /**
 * Check if there is any payments/checkouts on the order that we can use. 
 * Create new if there isn't.
 *
 * @order \Drupal\commerce_order\Entity\OrderInterface $order
 *   The complete order object
 *
 * @return
 * /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment
 */
  public function getPaymentFromOrder($order) {
    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payments */
    $payments = \Drupal::entityTypeManager()
      ->getStorage('commerce_payment')
      ->loadMultipleByOrder($order);
    if (count($payments) == 0 ) {
      // no payments on order.
      $paysonHandler = new PaysonHandler($this->payment_gateway_plugin);
      return $paysonHandler->connectPayson($this->checkoutPostData($order));
    }
    foreach($payments AS $row => $payment) {
      $paysonHandler = new PaysonHandler($payment->getPaymentGateway()->getPlugin());

      // get checkout from payson and check the state.
      $paysonCheckout = $paysonHandler->getCheckout($payment->getRemoteId());
      if($paysonCheckout ) {
          switch ($paysonCheckout->status) {
              case 'created':
                $cartUnchanged = false;
                $orderItems = $this->orderItems($order);
                // check payment against order.
                if ( (round($order->getTotalPrice()->getNumber()) == $paysonCheckout->order->totalPriceIncludingTax) &&
                  ( count($paysonCheckout->order->items) == count($orderItems)) ) {
                  // amount is the same and number of rows are the same.
                  foreach ($orderItems as $row => $item) {
                    // check that all the lines are the same.
                    if ($item['reference'] == $paysonCheckout->order->items[$row]->reference &&
                        round($item['quantity']) == $paysonCheckout->order->items[$row]->quantity){
                        // same product SKU and same quantity.
                        $cartUnchanged = true;
                        return $paysonCheckout;
                    }else {
                        $cartUnchanged = false;
                    }
                  }
                }

                break;
              case 'expired':
                // payment has expired, delete it.
                // they should usually be deleted from the IPN notification.
                $payment->delete();
                // and create new.
                return $paysonHandler->connectPayson($this->checkoutPostData($order));
                break;
              case 'readyToShip':
                $payment->setState($this->configuration['order_state']);
                $payment->save();
                // the event handler should send the user to the correct page here.
                // return the checkout.
                $cartUnchanged = true;
                break;
              case 'shipped':
              case 'paidToAccount':
                $payment->setState('completed');
                $payment->save();
                // the event handler should send the user to the correct page here.
                // return the checkout.
                $cartUnchanged = true;
                break;
              case 'canceled':
                // Checkout is cancelled, delete and create new.
                $payment->delete();
                // and create new.
                return $paysonHandler->connectPayson($this->checkoutPostData($order));
                break;
          }
      }
    }
    if (!$cartUnchanged) {
      // we have checked all the checkouts on the order and none are valid. Create new.
      return $paysonHandler->connectPayson($this->checkoutPostData($order));
    }
  }
  
}
